"""Base Module Generator.

This module is responsible for GeneratorAbstract and ModuleGenerator.

- Author: Jongkuk Lim
- Contact: lim.jeikei@gmail.com
"""
import inspect
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
from torch import nn

from kindle.utils.torch_utils import make_divisible


class GeneratorAbstract(ABC):
    """Abstract Module Generator."""

    CHANNEL_DIVISOR: int = 8

    def __init__(
        self,
        *args,
        keyword_args: Optional[Dict[str, Any]] = None,
        from_idx: Union[int, List[int]] = -1,
        in_channels: Tuple[int] = (0,),
        width_multiply: float = 1.0,
    ):
        """Initialize module generator.

        Args:
            *args: Module arguments
            from_idx: Module input index
            in_channels: Number of input channel
            width_multiply: Channel width multiply
        """
        self.args = tuple(args)
        self.keyword_args = keyword_args
        self.from_idx = from_idx
        self.in_channels = in_channels
        self.width_multiply = width_multiply

    @property
    def name(self) -> str:
        """Module name."""
        return self.__class__.__name__.replace("Generator", "")

    def _get_module(self, module: Union[nn.Module, List[nn.Module]]) -> nn.Module:
        """Get module from __call__ function."""
        if isinstance(module, list):
            module = nn.Sequential(*module)

        # error: Incompatible types in assignment (expression has type "Union[Tensor, Module, int]",
        # variable has type "Union[Tensor, Module]")
        # error: List comprehension has incompatible type List[int];
        # expected List[Union[Tensor, Module]]
        module.n_params = sum([x.numel() for x in module.parameters()])  # type: ignore
        # error: Cannot assign to a method
        module.name = self.name  # type: ignore

        return module

    @classmethod
    def _get_divisible_channel(cls, n_channel: int) -> int:
        """Get divisible channel by default divisor.

        Args:
            n_channel: number of channel.

        Returns:
            Ex) given {n_channel} is 52 and {GeneratorAbstract.CHANNEL_DIVISOR} is 8.,
                return channel is 56 since ceil(52/8) = 7 and 7*8 = 56
        """
        return make_divisible(n_channel, divisor=cls.CHANNEL_DIVISOR)

    def _get_kwargs(
        self, module_type: Any, args: Union[Tuple[Any, ...], List[Any]]
    ) -> Dict[str, Any]:
        """Get keyword argument with default argument values.

        Args:
            module_type: module class. Ex) nn.Conv
            args: ordered argument values

        Returns:
            converted keyword argument dictionary
            from the ordered argument values.
        """
        arg_spec = inspect.getfullargspec(module_type)
        kwarg_names = arg_spec.args
        if "self" in kwarg_names:
            kwarg_names.remove("self")

        if arg_spec.defaults is None:  # No keyword arguments
            arg_values = [
                None,
            ] * len(kwarg_names)
        else:
            arg_values = [
                None,
            ] * (len(kwarg_names) - len(arg_spec.defaults))
            arg_values += list(arg_spec.defaults)
        arg_values[: len(args)] = args

        kwarg_dict = dict(zip(kwarg_names, arg_values))
        if self.keyword_args is not None:
            for key, val in self.keyword_args.items():
                kwarg_dict[key] = val

        return kwarg_dict

    @property
    @abstractmethod
    def out_channel(self) -> Union[int, List[int]]:
        """Out channel of the module."""

    @property
    @abstractmethod
    def in_channel(self) -> Union[int, List[int]]:
        """In channel of the module."""

    @abstractmethod
    def compute_out_shape(
        self, size: np.ndarray, repeat: int = 1
    ) -> Union[List[int], List[List[int]]]:
        """Compute output shape when {size} is given.

        Args: input size to compute output shape.
        """

    @abstractmethod
    def __call__(self, repeat: int = 1) -> nn.Module:
        """Returns nn.Module component."""

    @property
    @abstractmethod
    def kwargs(self) -> Dict[str, Any]:
        """Generate keyword argument of the module."""


class ModuleGenerator:
    """Module generator class."""

    def __init__(
        self, module_name: str, custom_module_paths: Optional[Union[str, List]] = None
    ):
        """Generate module based on the {module_name}

        Args:
            module_name: {module_name}Generator class must have been implemented.
            custom_module_paths: paths to find custom module generators.
                Default location to find module generator is 'kindle.generator'.
                If {custom_module_paths} is provided, ModuleGenerator will expand its search area.
        """
        self.module_name = module_name
        self.generator_paths = ["kindle.generator"]
        if custom_module_paths is not None:
            if isinstance(custom_module_paths, str):
                paths = [custom_module_paths]
            else:
                paths = custom_module_paths
            self.generator_paths += paths

    def __call__(self, *args, keyword_args: Optional[Dict[str, Any]] = None, **kwargs):
        if self.module_name.startswith("nn."):
            generator_name = "TorchNNModuleGenerator"
            kwargs["module_name"] = self.module_name
        else:
            generator_name = f"{self.module_name}Generator"

        kwargs["keyword_args"] = keyword_args

        for path in self.generator_paths:
            if hasattr(__import__(path, fromlist=[""]), generator_name):
                return getattr(__import__(path, fromlist=[""]), generator_name)(
                    *args, **kwargs
                )

        raise Exception(f"{generator_name} can not be found.")



"""Flatten module generator.

- Author: Jongkuk Lim
- Contact: lim.jeikei@gmail.com
"""
from typing import Any, Dict, List

import numpy as np
from torch import nn

from kindle.generator.base_generator import GeneratorAbstract


class FlattenGenerator(GeneratorAbstract):
    """Flatten module generator."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.in_shape = [
            -1,
        ]

    @property
    def out_channel(self) -> int:
        return np.prod(self.in_shape)

    @property
    def in_channel(self) -> int:
        return -1

    @property
    def kwargs(self) -> Dict[str, Any]:
        kwargs = self._get_kwargs(nn.Flatten, self.args)
        return kwargs

    def compute_out_shape(self, size: np.ndarray, repeat: int = 1) -> List[int]:
        return [self.out_channel]

    def __call__(self, repeat: int = 1):
        return self._get_module(nn.Flatten(**self.kwargs))



"""Conv module generator.

- Author: Jongkuk Lim
- Contact: lim.jeikei@gmail.com
"""
from typing import Any, Dict, List, Union

import numpy as np
import torch
from torch import nn

from kindle.generator.base_generator import GeneratorAbstract


class ConvGenerator(GeneratorAbstract):
    """Conv2d generator for parsing module."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def out_channel(self) -> int:
        """Get out channel size."""
        return self._get_divisible_channel(self.args[0] * self.width_multiply)

    @property
    def in_channel(self) -> int:
        """Get in channel size."""
        # error: Value of type "Optional[List[int]]" is not indexable
        return self.in_channels[self.from_idx]  # type: ignore

    @property
    def base_module(self) -> nn.Module:
        """Returns module class from kindle.common_modules based on the class name."""
        return getattr(__import__("kindle.modules", fromlist=[""]), self.name)

    @property
    def kwargs(self) -> Dict[str, Any]:
        args = [self.in_channel, self.out_channel, *self.args[1:]]
        kwargs = self._get_kwargs(self.base_module, args)

        return kwargs

    @torch.no_grad()
    def compute_out_shape(
        self, size: Union[list, np.ndarray], repeat: int = 1
    ) -> List[int]:
        module: nn.Module = self(repeat=repeat)
        module.eval()
        module_out: torch.Tensor = module(torch.zeros([1, *list(size)]))
        return list(module_out.shape[-3:])

    def __call__(self, repeat: int = 1):
        kwargs = self.kwargs

        if repeat > 1:
            stride = 1
            # Important!: stride only applies at the end of the repeat.
            if kwargs["stride"] > 1:
                stride = kwargs["stride"]
                kwargs["stride"] = 1

            module = []
            for i in range(repeat):
                if stride > 1 and i == repeat - 1:
                    kwargs["stride"] = stride

                module.append(self.base_module(**kwargs))
                kwargs["in_channels"] = self.out_channel
        else:
            module = self.base_module(**kwargs)

        return self._get_module(module)


class DWConvGenerator(ConvGenerator):
    """Depth-wise convolution generator for parsing module."""


class FocusGenerator(ConvGenerator):
    """Focus convolution generator."""

    @property
    def in_channel(self) -> int:
        """Get in channel size."""
        return self.in_channels[self.from_idx] * 4  # type: ignore



"""Module Description.

- Author: Jongkuk Lim
- Contact: lim.jeikei@gmail.com
"""
import ast
from typing import Union

import torch
import torch.nn.functional as F
from torch import nn


class Activation:
    """Convert string activation name to the activation class."""

    def __init__(self, act_type: Union[str, None]) -> None:
        """Convert string activation name to the activation class.

        Args:
            act_type: Activation name.

        Returns:
            nn.Identity if {type} is None.
        """
        self.name = act_type
        self.args = [1] if self.name == "Softmax" else []

    def __call__(self) -> nn.Module:
        if self.name is None:
            return nn.Identity()
        if hasattr(nn, self.name):
            return getattr(nn, self.name)(*self.args)

        return ast.literal_eval(self.name)()


class SiLU(nn.Module):
    """Export-friendly version of nn.SiLU()

    SiLU https://arxiv.org/pdf/1606.08415.pdf
    """

    @staticmethod
    def forward(x: torch.Tensor) -> torch.Tensor:
        """Forward SiLU activation."""
        return x * torch.sigmoid(x)


class Hardswish(nn.Module):
    """Export-friendly version of nn.Hardswish()"""

    @staticmethod
    def forward(x: torch.Tensor) -> torch.Tensor:
        """Forward Hardswish activation."""
        return x * F.hardtanh(x + 3, 0.0, 6.0) / 6.0


class Mish(nn.Module):
    """Mish https://github.com/digantamisra98/Mish."""

    @staticmethod
    def forward(x: torch.Tensor) -> torch.Tensor:
        """Forward Mish activation."""
        return x * F.softplus(x).tanh()
